"""Console script for oceantide."""
import sys
import click


@click.command()
def main(args=None):
    """Console script for oceantide."""
    click.echo("Replace this message by putting your code into " "oceantide.cli.main")
    click.echo("See click documentation at http://click.pocoo.org/")
    return 0
